Public Delegate Function MyFxDelegate(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double) As Double
Public Delegate Function SayFxDelegate() As String

Public Class CConjugateGradient1
  Dim m_MyFx As MyFxDelegate


  Function MyFxEx(ByVal nNumVars As Integer, _
                ByRef fX() As Double, ByRef fParam() As Double, _
                ByRef fDeltaX() As Double, ByVal fLambda As Double) As Double
    Dim I As Integer
    Dim fXX(nNumVars) As Double

    For I = 0 To nNumVars - 1
      fXX(I) = fX(I) + fLambda * fDeltaX(I)
    Next I

    MyFxEx = m_MyFx(nNumVars, fXX, fParam)
  End Function

  Private Sub GetGradients( _
       ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
       ByRef fDeriv() As Double, ByRef fDerivNorm As Double)

    Dim I As Integer
    Dim fXX, H, Fp, Fm As Double

    fDerivNorm = 0
    For I = 0 To nNumVars - 1
      fXX = fX(I)
      H = 0.01 * (1 + Math.Abs(fXX))
      fX(I) = fXX + H
      Fp = m_MyFx(nNumVars, fX, fParam)
      fX(I) = fXX - H
      Fm = m_MyFx(nNumVars, fX, fParam)
      fX(I) = fXX
      fDeriv(I) = (Fp - Fm) / 2 / H
      fDerivNorm += fDeriv(I) ^ 2
    Next I
    fDerivNorm = Math.Sqrt(fDerivNorm)
  End Sub

  Function LinSearch_DirectSearch(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
            ByRef fLambda As Double, ByRef fDeltaX() As Double, ByVal InitStep As Double, _
            ByVal MinStep As Double) As Boolean
    Dim F1, F2 As Double

    F1 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda)

    Do
      F2 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda + InitStep)
      If F2 < F1 Then
        F1 = F2
        fLambda += InitStep
      Else
        F2 = MyFxEx(nNumVars, fX, fParam, fDeltaX, fLambda - InitStep)
        If F2 < F1 Then
          F1 = F2
          fLambda -= InitStep
        Else
          ' reduce search step size
          InitStep /= 10
        End If
      End If
    Loop Until InitStep < MinStep

    Return True

  End Function


  Public Function CalcOptim(ByVal nNumVars As Integer, ByRef fX() As Double, ByRef fParam() As Double, _
                    ByVal fEpsFx As Double, ByVal nMaxIter As Integer, _
                    ByRef nIter As Integer, ByRef sErrorMsg As String, _
                    ByVal MyFx As MyFxDelegate) As Double

    Dim I As Integer
    Dim fDeriv(nNumVars) As Double, fDerivOld(nNumVars) As Double
    Dim F, fDFNorm, fDFNormOld As Double
    Dim fLambda, fLastF As Double

    m_MyFx = MyFx

    ' calculate and function value at initial point
    fLastF = MyFx(nNumVars, fX, fParam)

    GetGradients(nNumVars, fX, fParam, fDeriv, fDFNorm)

    fLambda = 0
    If LinSearch_DirectSearch(nNumVars, fX, fParam, fLambda, fDeriv, 0.1, 0.000001) Then
      For I = 0 To nNumVars - 1
        fX(I) += fLambda * fDeriv(I)
      Next I
    Else
      sErrorMsg = "Failed linear search"
      Return fLastF
    End If

    nIter = 1
    Do
      nIter += 1
      If nIter > nMaxIter Then
        sErrorMsg = "Reached maximum iterations limit"
        Exit Do
      End If
      fDFNormOld = fDFNorm
      For I = 0 To nNumVars - 1
        fDerivOld(I) = fDeriv(I) ' save old gradient
      Next I
      GetGradients(nNumVars, fX, fParam, fDeriv, fDFNorm)
      For I = 0 To nNumVars - 1
        fDeriv(I) = (fDFNorm / fDFNormOld) ^ 2 * fDerivOld(I) - fDeriv(I)
      Next I
      If fDFNorm <= fEpsFx Then
        sErrorMsg = "Gradient norm meets convergence criteria"
        Exit Do
      End If
      '    For I = 0 To nNumVars - 1
      '      fDeriv(I) = -fDeriv(I) / fDFNorm
      '    Next I
      fLambda = 0
      '    If LinSearch_Newton(fX, nNumVars, fLambda, fDeriv, 0.0001, 100) Then
      If LinSearch_DirectSearch(nNumVars, fX, fParam, fLambda, fDeriv, 0.1, 0.000001) Then
        For I = 0 To nNumVars - 1
          fX(I) += fLambda * fDeriv(I)
        Next I
        F = MyFx(nNumVars, fX, fParam)
        If Math.Abs(F - fLastF) < fEpsFx Then
          sErrorMsg = "Successive function values meet convergence criteria"
          Exit Do
        Else
          fLastF = F
        End If

      Else
        sErrorMsg = "Failed linear search"
        Exit Do
      End If
    Loop

    Return fLastF

  End Function


End Class
